% ICML2011
%
% Runs synthetic experiments described in
% 
%    "Max-margin Learning for Lower Linear Envelope Potentials in Binary
%    Markov Random Fields", Stephen Gould, ICML 2011.
%
% Copyright (c) 2011, Stephen Gould <stephen.gould@anu.edu.au>
% All rights reserved.
%

rand('state', 0);       % initialize random number generator

options.M = 8;          % size of checkerboard M-by-M
options.N = 16;         % number of pixels per square (clique) N-by-N
options.K = 10;         % number of lower linear functions
options.cliques = [];   % mapping of variables to higher-order cliques MN-by-MN
options.figWnd = -1;    % figure handle (or -1 for no visualization)
options.maxIters = 100; % maximum learning iterations
options.eps = 1.0e-16;  % constraint violation threshold
options.bPairwise = 0;  % include pairwise weight

% create checkerboard data
y = zeros(options.M * options.N, options.M * options.N);
options.cliques = zeros(options.M * options.N, options.M * options.N);
for i = 1:options.M,
    for j = 1:options.M,
        rowIndx = (1:options.N) + (i-1)*options.N;
        colIndx = (1:options.N) + (j-1)*options.N;
        y(rowIndx, colIndx) = mod(i + j, 2);
        options.cliques(rowIndx, colIndx) = options.M * (j - 1) + i;
    end;
end;

% create noisey observations
eta = [0.1, 0.1];       % signal-to-noise ratio for black and white pixels
x = 2 * (rand(size(y)) - 0.5) + eta(1) * (1 - y) - eta(2) * y;
eta2 = [0.5, 0.1];      % signal-to-noise ratio for black and white pixels
x2 = 2 * (rand(size(y)) - 0.5) + eta2(1) * (1 - y) - eta2(2) * y;
    
% learn pairwise crf
options.figWnd = -1;
[~, history] = pairwiseLearn(x, y, options);
y_bestPairwise = history(end).y_hat;

% learn higher-order crf
options.figWnd = figure;
[parameters, history] = linEnvLearn(x, y, options);

% learn pairwise crf
options.figWnd = -1;
[~, history2] = pairwiseLearn(x2, y, options);
y_bestPairwise2 = history2(end).y_hat;

% learn higher-order crf
options.figWnd = figure;
[parameters2, history2] = linEnvLearn(x2, y, options);

% plot results
figure; colormap(gray);
subplot(2, 5, 1); image(255 * y); axis off; title('ground-truth');
subplot(2, 5, 2); imagesc(x); axis off; title('data');
subplot(2, 5, 3); imagesc(255 * y_bestPairwise); axis off; title('pairwise');
subplot(2, 5, 4); imagesc(255 * history(3).y_hat); axis off; title('iter. 3');
subplot(2, 5, 5); imagesc(255 * history(end).y_hat); axis off; title('last iter.');

subplot(2, 5, 6); image(255 * y); axis off;
subplot(2, 5, 7); imagesc(x2); axis off;
subplot(2, 5, 8); imagesc(255 * y_bestPairwise2); axis off;
subplot(2, 5, 9); imagesc(255 * history2(3).y_hat); axis off;
subplot(2, 5, 10); imagesc(255 * history2(end).y_hat); axis off;

figure;
theta = [parameters.linEnvCoeffs(1, 2); parameters.linEnvCoeffs(:, 2)] + ...
    [0; parameters.linEnvCoeffs(:, 1)] .* linspace(0, options.N * options.N, options.K + 1)';
plot(theta / parameters.unaryWeight, 'bo-', 'LineWidth', 4);
hold on;
theta = [parameters2.linEnvCoeffs(1, 2); parameters2.linEnvCoeffs(:, 2)] + ...
    [0; parameters2.linEnvCoeffs(:, 1)] .* linspace(0, options.N * options.N, options.K + 1)';
plot(theta / parameters2.unaryWeight, 'rd-', 'LineWidth', 4);
hold off;
legend('\eta = [0.1, 0.1]', '\eta = [0.5, 0.1]');

grid on;
title('Learned lower linear envelopes');
