/*****************************************************************************
** This is a reference implementation for the synthetic experiments on lower
** linear envelope inference and learning described in
**
**   "Max-margin Learning for Lower Linear Envelope Potentials in Binary
**    Markov Random Fields", Stephen Gould, ICML 2011.
**
** Copyright (c) 2011, Stephen Gould <stephen.gould@anu.edu.au>
** All rights reserved.
**
*****************************************************************************/

// c++ standard headers
#include <cstdlib>
#include <cstdio>
#include <cmath>
#include <limits>
#include <vector>
#include <list>
#include <set>

// matlab
#include "mex.h"
#include "matrix.h"

// kolmogorov's maxflow-v3.01
#include "maxflow-v3.01.src/graph.h"

using namespace std;

// usage --------------------------------------------------------------------

void usage()
{
    mexPrintf("LINENVINF    Lower Linear Envelope Inference\n");
    mexPrintf("\n");
    mexPrintf("USAGE: y = linEnvInf(unaryWeights, pairwiseWeight, linEnvCoeffs);\n");
    mexPrintf("\n");
}

// mexFunction --------------------------------------------------------------

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    // constants
    const int M = 8;   // size of checkerboard M-byM
    const int N = 16;  // number of pixels per square N-by-N

    // check number of input and output arguments
    if ((nrhs != 3) || (nlhs != 1)) {
        usage();
        return;
    }

    const int nVariables = mxGetNumberOfElements(prhs[0]);
    mxAssert(nVariables != 0, "no variables");
    mxAssert((mxGetM(prhs[0]) == M * N) && (mxGetN(prhs[0]) == N * M),
        "expecting NM-by-NM unaryWeights matrix");

    // construct s-t graph
    typedef Graph<double, double, double> GraphType;
    GraphType *g = new GraphType(nVariables, 8 * nVariables);

    g->add_node(nVariables);

    // add unary terms
    const double *p = mxGetPr(prhs[0]);
    for (int i = 0; i < nVariables; i++) {
        g->add_tweights(i, 0.0, p[i]);
    }

    // add pairwise terms
    mxAssert(mxGetNumberOfElements(prhs[1]) == 1, "expecting scalar for pairwiseWeight");
    double pairwiseWeight = mxGetScalar(prhs[1]);
    mxAssert(pairwiseWeight >= 0.0, "pairwiseWeight must be non-negative");

    if (pairwiseWeight > 0.0) {
        // vertical edges
        for (int y = 1; y < M * N; y++) {
            for (int x = 0; x < M * N; x++) {
                g->add_edge(M * N * x + y, M * N * x + y - 1, pairwiseWeight, pairwiseWeight);
            }
        }

        // horizontal edges
        for (int y = 0; y < M * N; y++) {
            for (int x = 1; x < M * N; x++) {
                g->add_edge(M * N * x + y, M * N * (x - 1) + y, pairwiseWeight, pairwiseWeight);
            }
        }        
    }

    // add higher-order term for each clique
    const int K = mxGetM(prhs[2]);
    if (K != 0) {
        mxAssert(mxGetN(prhs[2]) == 2, "expecting Kx2 matrix for linEnvCoeffs");
        p = mxGetPr(prhs[2]);

        // iterate over cliques
        for (int y = 0; y < M * N; y += N) {
            for (int x = 0; x < M * N; x += N) {

                // add higher-order term
                double a = p[0];
                for (int i = 0; i < N; i++) {
                    for (int j = 0; j < N; j++) {
                        g->add_tweights(M * N * (x + j) + (y + i), 0.0, a);
                    }
                }

                for (int k = 1; k < K; k++) {
                    double da = p[k] - p[k - 1];
                    double db = p[K + k] - p[K + k - 1];
                    if (fabs(da) < 1.0e-6) continue;
                    mxAssert((da < 0.0) && (db > 0.0), "invalid linear envelope coefficients");

                    int z = g->add_node(1);
                    g->add_tweights(z, -1.0 * N * N * da, db);

                    for (int i = 0; i < N; i++) {
                        for (int j = 0; j < N; j++) {
                            g->add_edge(M * N * (x + j) + (y + i), z, 0.0, -da);
                        }
                    }

                    // constraint on z
                    if (k != 1) {
                        g->add_edge(z, z - 1, numeric_limits<double>::max(), 0.0);
                    }
                }
            }
        }
    }

    // find min-st-cut
    double e = g->maxflow();
    mexPrintf("min-cut has value %f\n", e);

    // decode solution
    plhs[0] = mxCreateDoubleMatrix(mxGetM(prhs[0]), mxGetN(prhs[0]), mxREAL);
    double *y = mxGetPr(plhs[0]);
    for (int i = 0; i < nVariables; i++) {
        y[i] = (g->what_segment(i) == GraphType::SOURCE) ? 1.0 : 0.0;
    }

    // free graph
    delete g;
}
