% LINENVLEARN   Lower linear enevelope max-margin learning
%
% This is a reference implementation for the synthetic experiments on lower
% linear enevelope inference and learning described in
%
%    "Max-margin Learning for Lower Linear Envelope Potentials in Binary
%    Markov Random Fields", Stephen Gould, ICML 2011.
%
% Copyright (c) 2011, Stephen Gould <stephen.gould@anu.edu.au>
% All rights reserved.
%

function [params, history] = linEnvLearn(x, y, options);

% true feature vector
truePhi = featureVector(x, y, options);

% initialize parameters
params.unaryWeight = 1.0;
params.pairwiseWeight = 0.0;
params.linEnvCoeffs = zeros(options.K, 2);

numVariables = numel(y);
numParameters = options.K + 2; % unary plus linear envelope coefficients
if (options.bPairwise == 1), numParameters = numParameters + 1; end;

% construct QP objective
P = eye(numParameters + 1, numParameters + 1);
P(end, end) = 0.0;
q = zeros(size(P, 1), 1);
q(end) = 1.0;

% construct non-decreasing concave constraint
Dxx = toeplitz([-1 zeros(1, options.K - 2)], [-1 2 -1 zeros(1, options.K - 2)]);
G = [zeros(size(Dxx, 1), numParameters - options.K - 1), Dxx, zeros(size(Dxx, 1), 1)];
h = zeros(size(Dxx, 1), 1);

% positivity constraint on pairwise weight
if (options.bPairwise == 1),
    G = [G; 0, 1, zeros(1, size(G, 2) - 2)];
    h = [h; 0];
end;

% positivity constraint on slack
G = [G; zeros(1, size(G, 2) - 1), 1];
h = [h; 0];

% iterate until convergence
for t = 1:options.maxIters,
    % solve max-margin qp
    warning off;
    theta = quadprog(P, q, -1.0 * G, -1.0 * h);
    warning on;
    
    % decode parameters
    params.unaryWeight = theta(1);
    if (options.bPairwise == 1), 
        params.pairwiseWeight = max([0.0, theta(2)]);
        for k = 1:options.K,
            params.linEnvCoeffs(k, 1) = theta(k + 3) - theta(k + 2);
            params.linEnvCoeffs(k, 2) = theta(k + 3) - params.linEnvCoeffs(k, 1) * k;
        end;        
    else
        for k = 1:options.K,
            params.linEnvCoeffs(k, 1) = theta(k + 2) - theta(k + 1);
            params.linEnvCoeffs(k, 2) = theta(k + 2) - params.linEnvCoeffs(k, 1) * k;
        end;
    end;
    params.linEnvCoeffs(:, 1) = options.K / (options.N * options.N) * params.linEnvCoeffs(:, 1);
        
    % infer solution
    y_hat = linEnvInf(params.unaryWeight * x, params.pairwiseWeight, ...
        params.linEnvCoeffs);

    % update history
    history(t) = struct('params', params, 'y_hat', y_hat);
    
    % show results for this iteration
    if (options.figWnd > 0),
        figure(options.figWnd);
        subplot(1, 2, 1);
        plot(theta(end - options.K - 1:end - 1) / abs(theta(1)), 'lineWidth', 2);
        a = axis; axis([1, options.K + 1, a(3), a(4)]); grid on;
        title('linear envelope');
        subplot(1, 2, 2); colormap(gray); image(255 * y_hat);
        title('inferred labeling');
        drawnow;
    end;
    
    % infer most violated constraint
    y_loss = linEnvInf(params.unaryWeight * x + 2 * (y - 0.5) / numVariables, ...
        params.pairwiseWeight, params.linEnvCoeffs);
    
    % add constraint
    phi = featureVector(x, y_loss, options);
    loss = sum(y_loss(:) ~= y(:)) / numVariables;
    slack = loss - (phi - truePhi) * theta(1:end - 1);
    violation = slack - theta(end);
    disp(['iter = ', int2str(t), ...
        ', loss = ', num2str(sum(y_hat(:) ~= y(:)) / numVariables), ...
        ', max-violation = ', num2str(violation)]);
        
    % check for convergence
    if (violation < options.eps), disp('converged'); break; end;
    
    G = [G; phi - truePhi, 1];
    h = [h; loss];
end;

% feature vector --------------------------------------------------------

function [phi] = featureVector(x, y_hat, options);

unaryPhi = sum(x(y_hat == 1));
pairwisePhi = sum(sum(abs(diff(y_hat)))) + sum(sum(abs(diff(y_hat'))));
highOrderPhi = zeros(1, options.K + 1);
for m = 1:options.M * options.M,
    p = sum(y_hat(options.cliques == m)) / (options.N * options.N);
    k = floor(p * options.K);
    highOrderPhi(k + 1) = highOrderPhi(k + 1) + k - p * options.K + 1;
    if (k < options.K),
        highOrderPhi(k + 2) = highOrderPhi(k + 2) + p * options.K - k;
    end;
end;

if (options.bPairwise == 1),
    phi = [unaryPhi, pairwisePhi, highOrderPhi];
else
    phi = [unaryPhi, highOrderPhi];
end;

return;
